/* functions for dealing with a tandem bot */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include "eggdrop.h"
#include "tandem.h"

extern tand_t tand[];
extern int tands;
extern char origbotname[];
extern int dcc_total;
extern struct dcc_t dcc[];
extern int isolate;
extern int noshare;
extern int passive;
extern int share_users;
extern int nosave;
extern int periodic_save;

char *cursend();
void *dup_userlist();

char *attrname(a)
int a;
{
  char s[21];
  switch(a) {
  case USER_OP: strcpy(s,"op"); break;
  case USER_DEOP: strcpy(s,"deop"); break;
  case USER_BAN: strcpy(s,"ban"); break;
  case USER_KICK: strcpy(s,"kick"); break;
  case USER_MASTER: strcpy(s,"master"); break;
  case USER_FRIEND: strcpy(s,"friend"); break;
  case USER_XFER: strcpy(s,"xfer"); break;
  case USER_IGNORE: strcpy(s,"ignore"); break;
  case USER_TANDEM: strcpy(s,"tandem"); break;
  case USER_PARTY: strcpy(s,"party/share"); break;
  case USER_COMMON: strcpy(s,"common"); break;
  case BOT_AUTO: strcpy(s,"auto"); break;
  case BOT_LEAF: strcpy(s,"leaf"); break;
  case BOT_REJECT: strcpy(s,"reject"); break;
  default: strcpy(s,"???(upgrade!)"); break;
  }
  return s;
}

void tand_chat(idx,z,par)
int idx,z; char *par;
{
  char nk[40],from[40],s[512]; int i;
  nsplit(from,par); if (strchr(from,'@')==NULL) {
    /* notice from a bot */
    if ((strcasecmp(from,origbotname)==0) || (!in_chain(from))) {
      /* from me?!  -or-  from nonexistant bot?! */
      /* don't pass it on -- instead, warn others */
      tprintf(z,"chat %s WARNING: FAKE BOT NOTICES\n",origbotname);
      return;
    }
    sprintf(s,"*** (%s) %s",from,par);
  }
  else sprintf(s,"<%s> %s",from,par);
  /* make sure the bot is valid */
  splitc(nk,from,'@');
  if ((strcasecmp(from,origbotname)==0) || (!in_chain(from))) {
    tprintf(z,"chat %s WARNING: FAKE BOT NOTICES\n",origbotname);
    return;
  }
  if (!isolate) for (i=0; i<dcc_total; i++)
    if ((dcc[i].type==DCC_CHAT) && (!(dcc[i].u.chat->status&STAT_BLIND)))
      tprintf(dcc[i].sock,"%s\n",s);
  if (nk[0]) tandout_but(idx,"chat %s@%s %s\n",nk,from,par);
  else tandout_but(idx,"chat %s %s\n",from,par);
}

void tand_act(idx,z,par)
int idx,z; char *par;
{
  char nk[40],from[40]; int i;
  nsplit(from,par); splitc(nk,from,'@');
  if ((!nk[0]) || (strcasecmp(from,origbotname)==0) || (!in_chain(from))) {
    tprintf(z,"chat %s WARNING: FAKE BOT NOTICES\n",origbotname);
    return;
  }
  if (!isolate) for (i=0; i<dcc_total; i++)
    if ((dcc[i].type==DCC_CHAT) && (!(dcc[i].u.chat->status&STAT_BLIND)))
      tprintf(dcc[i].sock,"* %s@%s %s\n",nk,from,par);
  tandout_but(idx,"act %s@%s %s\n",nk,from,par);
}

void tand_priv(idx,z,par)
int idx,z; char *par;
{
  char from[40],frombot[40],nk[40],to[40],tobot[40]; int i;
  nsplit(from,par); nsplit(tobot,par); splitc(to,tobot,'@');
  strcpy(frombot,from); splitc(nk,frombot,'@');
  if ((!to[0]) || (strcasecmp(frombot,origbotname)==0) ||
      (!in_chain(frombot))) {
    tprintf(z,"chat %s WARNING: FAKE BOT NOTICES\n",origbotname);
    return;
  }
  if (!to[0]) { strcpy(to,tobot); tobot[0]=0; }
  if (strcasecmp(tobot,origbotname)==0) {  /* for me! */
    if (strchr(from,'@')==NULL) add_note(to,from,par,-2);
    else {
      i=add_note(to,from,par,-1);
      if (i==0)  /* failed */
	tprintf(z,"priv %s %s No such user %s.\n",origbotname,from,to);
      else if (i==2)   /* stored */
	tprintf(z,"priv %s %s Not online; note stored.\n",origbotname,from);
      else    /* success! */
	tprintf(z,"priv %s %s Note sent to %s.\n",origbotname,from,to);
    }
  }
  else {   /* pass it on */
    i=nextbot(tobot); if (i<0)
      tandout_but(idx,"priv %s %s@%s %s\n",from,to,tobot,par);
    else tprintf(dcc[i].sock,"priv %s %s@%s %s\n",from,to,tobot,par);
  }
}

void tand_bye(idx,z,par)
int idx,z; char *par;
{
  log(LOG_MISC,"Disconnected from: %s",dcc[idx].nick);
  tandout_but(idx,"chat %s Disconnected from: %s\n",origbotname,
	      dcc[idx].nick);
  tandout_but(idx,"unlinked %s\n",dcc[idx].nick);
  cancel_user_xfer(idx);
  rembot(dcc[idx].nick,dcc[idx].nick); unvia(idx,dcc[idx].nick);
  chatout("*** Disconnected from: %s\n",dcc[idx].nick);
  tprintf(z,"*bye\n"); close(z); shutdown(z,2); lostdcc(idx);
}

void tand_who(idx,z,par)
int idx,z; char *par;
{
  char *p; char from[40],new[40]; int i;
  nsplit(from,par); p=strchr(from,'@'); if (p==NULL) {
    sprintf(new,"%s@%s",from,dcc[idx].nick); strcpy(from,new);
  }
  if (strcasecmp(par,origbotname)==0) par[0]=0;  /* (for me) */
  if (par[0]) {     /* pass it on */
    i=nextbot(par); if (i<0)
      tandout_but(idx,"who %s %s\n",from,par);
    else tprintf(dcc[i].sock,"who %s %s\n",from,par);
  }
  else {
    log(LOG_CMDS,"#%s# who",from);
    tell_tandem_who(z,from);
  }
}

void tand_version(idx,z,par)
int idx,z; char *par;
{
  strcpy(dcc[idx].u.tand->version,par);
  if ((share_users) && (match_attr_handle(dcc[idx].nick,BOT_SHARE))) {
    if (passive) {
      if (dcc[idx].u.tand->status&STAT_CALLED) {
	if (can_resync(dcc[idx].nick)) tprintf(z,"resync?\n");
	else tprintf(z,"userfile?\n");
	dcc[idx].u.tand->status|=STAT_OFFERED;
      }
    }
    else {
      if (can_resync(dcc[idx].nick)) tprintf(z,"resync?\n");
      else tprintf(z,"userfile?\n");
      dcc[idx].u.tand->status|=STAT_OFFERED;
    }
  }
}

void tand_looptest(idx,z,par)
int idx,z; char *par;
{
  log(LOG_MISC,"Ancient bot %s detected!  Disconnecting...",par);
  tprintf(z,"reject %s %s\n",origbotname,par);
}

void tand_whoq(idx,z,par)
int idx,z; char *par;
{
  /* who? <from@bot>   ->  send: whom <to@bot> <attr><nick> <bot> <host> */
  if (!isolate) log(LOG_CMDS,"#%s# whom",par);
  answer_whom(z,par);
  tandout_but(idx,"who? %s\n",par);
}

void tand_infoq(idx,z,par)
int idx,z; char *par;
{
  /* info? <from@bot>   -> send priv */
  log(LOG_CMDS,"#%s# botinfo",par);
  tell_tandem_info(z,par);
  tandout_but(idx,"info? %s\n",par);
}

void tand_whom(idx,z,par)
int idx,z; char *par;
{
  char to[40],tobot[40],nick[40],bot[40]; int i;
  nsplit(tobot,par); splitc(to,tobot,'@');
  if (strcasecmp(tobot,origbotname)==0) {
    nsplit(nick,par); nsplit(bot,par);
    if (nick[0]=='-') nick[0]=' ';
    for (i=0; i<dcc_total; i++) if (strcasecmp(dcc[i].nick,to)==0)
      tprintf(dcc[i].sock,"%-10s  %-9s  %s\n",nick,bot,par);
  }
  else {
    i=nextbot(tobot); if (i<0)
      tandout_but(idx,"whom %s@%s %s\n",to,tobot,par);
    else tprintf(dcc[i].sock,"whom %s@%s %s\n",to,tobot,par);
  }
}

void tand_ping(idx,z,par)
int idx,z; char *par;
{
  tprintf(dcc[idx].sock,"pong\n");
}

void tand_pong(idx,z,par)
int idx,z; char *par;
{
  dcc[idx].u.tand->status&=~STAT_PINGED;
}

void tand_link_cmd(idx,z,par)
int idx,z; char *par;
{
  char from[40],bot[40]; int i;
  /* link <from@bot> <who> <to-whom> */
  nsplit(from,par); nsplit(bot,par);
  if (strcasecmp(bot,origbotname)==0) {
    log(LOG_CMDS,"#%s# link %s",from,par);
    if (tandem_link(-2,par))
      tprintf(dcc[idx].sock,"priv %s %s Attempting link to %s ...\n",
	      origbotname,from,par);
    else tprintf(dcc[idx].sock,"priv %s %s Can't link there.\n",origbotname,
		 from);
  }
  else {
    i=nextbot(bot); if (i<0)
      tandout_but(idx,"link %s %s %s\n",from,bot,par);
    else tprintf(dcc[i].sock,"link %s %s %s\n",from,bot,par);
  }
}

void tand_unlink_cmd(idx,z,par)
int idx,z; char *par;
{
  char from[40],bot[40]; int i;
  nsplit(from,par); nsplit(bot,par);
  if (strcasecmp(bot,origbotname)==0) {
    log(LOG_CMDS,"#%s# unlink %s",from,par);
    if (tandem_unlink(-2,par))
      tprintf(dcc[idx].sock,"priv %s %s Unlinked %s.\n",origbotname,from,par);
    else tprintf(dcc[idx].sock,"priv %s %s Can't unlink %s.\n",origbotname,
		 from,par);
  }
  else {
    i=nextbot(bot); if (i<0)
      tandout_but(idx,"unlink %s %s %s\n",from,bot,par);
    else tprintf(dcc[i].sock,"unlink %s %s %s\n",from,bot,par);
  }
}

void tand_ufno(idx,z,par)
int idx,z; char *par;
{
  log(LOG_MISC,"User file rejected by %s: %s",dcc[idx].nick,par);
  dcc[idx].u.tand->status&=~STAT_OFFERED;
  if (!(dcc[idx].u.tand->status&STAT_GETTING)) 
    dcc[idx].u.tand->status&=~STAT_SHARE;
}

void tand_ufyes(idx,z,par)
int idx,z; char *par;
{
  if (dcc[idx].u.tand->status&STAT_OFFERED) {
    dcc[idx].u.tand->status&=~STAT_OFFERED;
    dcc[idx].u.tand->status|=STAT_SHARE;
    dcc[idx].u.tand->status|=STAT_SENDING;
    start_sending_users(z,idx);
    log(LOG_MISC,"Sending user file to %s",dcc[idx].nick);
  }
}

void tand_userfileq(idx,z,par)
int idx,z; char *par;
{
  int ok=1,i;
  flush_tbuf(dcc[idx].nick);
  if (!share_users) tprintf(z,"uf-no Not sharing userfile.\n");
  else if (!passive) tprintf(z,"uf-no Aggressive mode active.\n");
  else if (!match_attr_handle(dcc[idx].nick,BOT_SHARE))
    tprintf(z,"uf-no You are not +s for me.\n");
  else {
    for (i=0; i<dcc_total; i++) if (dcc[i].type==DCC_TANDEM)
      if ((dcc[i].u.tand->status&STAT_SHARE) &&
	  (dcc[i].u.tand->status&STAT_GETTING)) ok=0;
    if (!ok) tprintf(z,"uf-no Already downloading a userfile.\n");
    else {
      tprintf(z,"uf-yes\n"); dcc[idx].u.tand->status|=STAT_SHARE;
      clear_userlist2(dcc[idx].u.tand->user);
      /* copy just the tandem-bots into the new userlist (wanna keep those) */
      dcc[idx].u.tand->user=dup_userlist(1);
      dcc[idx].u.tand->status|=STAT_GETTING;
      log(LOG_MISC,"Downloading user file from %s",dcc[idx].nick);
    }
  }
}

void tand_resyncq(idx,z,par)
int idx,z; char *par;
{
  if (!share_users) tprintf(z,"resync-no Not sharing userfile.\n");
  else if (!match_attr_handle(dcc[idx].nick,BOT_SHARE))
    tprintf(z,"resync-no You are not +s for me.\n");
  else if (can_resync(dcc[idx].nick)) {
    tprintf(z,"resync!\n"); dump_resync(z,dcc[idx].nick);
    dcc[idx].u.tand->status&=~STAT_OFFERED;
    dcc[idx].u.tand->status|=STAT_SHARE;
    log(LOG_MISC,"Resync'd user file with %s",dcc[idx].nick);
  }
  else if (passive) {
    tprintf(z,"resync!\n");
    dcc[idx].u.tand->status&=~STAT_OFFERED;
    dcc[idx].u.tand->status|=STAT_SHARE;
    log(LOG_MISC,"Resyncing user file from %s",dcc[idx].nick);
  }
  else tprintf(z,"resync-no No resync buffer.\n");
}

void tand_resync(idx,z,par)
int idx,z; char *par;
{
  if (dcc[idx].u.tand->status&STAT_OFFERED) {
    if (can_resync(dcc[idx].nick)) {
      dump_resync(z,dcc[idx].nick);
      dcc[idx].u.tand->status&=~STAT_OFFERED;
      dcc[idx].u.tand->status|=STAT_SHARE;
      log(LOG_MISC,"Resync'd user file with %s",dcc[idx].nick);
    }
  }
}

void tand_resync_no(idx,z,par)
int idx,z; char *par;
{
  log(LOG_MISC,"Resync refused by %s: %s",dcc[idx].nick,par);
  flush_tbuf(dcc[idx].nick);
  tprintf(z,"userfile?\n");
}

#define CHKSEND if (!(dcc[idx].u.tand->status&STAT_SENDING)) return
#define CHKGET if (!(dcc[idx].u.tand->status&STAT_GETTING)) return
#define CHKSHARE if (!(dcc[idx].u.tand->status&STAT_SHARE)) return

void tand_ackuser(idx,z,par)
int idx,z; char *par;
{
  CHKSEND;
  if (strcasecmp(par,cursend(idx))==0) {
    if (send_next_user(z,idx))
      log(LOG_MISC,"Finished sending user list to %s.",dcc[idx].nick);
  }
  else {
    log(LOG_MISC,"* Desync sending user list to %s.",dcc[idx].nick);
    log(LOG_MISC,"* (Re-sending user '%s'.)",cursend(idx));
    resend_user(z,idx);
  }
}

void tand_end_userlist(idx,z,par)
int idx,z; char *par;
{
  CHKGET;
  log(LOG_MISC,"Finished receiving user list from %s.",dcc[idx].nick);
  if (!periodic_save) nosave=0;
  dcc[idx].u.tand->status&=~STAT_GETTING;
  use_new_buf(idx); write_userfile(); noshare=0;
}

void tand_user(idx,z,par)
int idx,z; char *par;
{
  CHKGET;
  store_new_user(idx,par);
}

void tand_email(idx,z,par)
int idx,z; char *par;
{
  CHKGET;
  set_handle_email2(dcc[idx].u.tand->user,cursend(idx),par);
}

void tand_dccdir(idx,z,par)
int idx,z; char *par;
{
  CHKGET;
  set_handle_dccdir2(dcc[idx].u.tand->user,cursend(idx),par);
}

void tand_comment(idx,z,par)
int idx,z; char *par;
{
  CHKGET;
  set_handle_comment2(dcc[idx].u.tand->user,cursend(idx),par);
}

void tand_info(idx,z,par)
int idx,z; char *par;
{
  CHKGET;
  set_handle_info2(dcc[idx].u.tand->user,cursend(idx),par);
}

void tand_laston(idx,z,par)
int idx,z; char *par;
{
  CHKGET;
  set_handle_laston2(dcc[idx].u.tand->user,cursend(idx),atol(par));
}

void tand_host(idx,z,par)
int idx,z; char *par;
{
  CHKGET;
  addhost_by_handle2(dcc[idx].u.tand->user,cursend(idx),par);
}

void tand_end_user(idx,z,par)
int idx,z; char *par;
{
  CHKGET;
  tprintf(z,"ackuser %s\n",cursend(idx)); noshare=0;
  dcc[idx].host[0]=0; if (!periodic_save) nosave=0;
}

void tand_chpass(idx,z,par)
int idx,z; char *par;
{
  char hand[40];
  CHKSHARE;
  shareout_but(idx,"chpass %s\n",par);
  noshare=1; nsplit(hand,par); change_pass_by_handle(hand,par); noshare=0;
  log(LOG_CMDS,"%s: newpass %s",dcc[idx].nick,hand);
}

void tand_chhand(idx,z,par)
int idx,z; char *par;
{
  char hand[40];
  CHKSHARE;
  shareout_but(idx,"chhand %s\n",par);
  noshare=1; nsplit(hand,par); change_handle(hand,par); noshare=0;
  log(LOG_CMDS,"%s: handle %s->%s",dcc[idx].nick,hand,par);
}

void tand_newuser(idx,z,par)
int idx,z; char *par;
{
  char etc[40],etc2[40],etc3[40];
  CHKSHARE;
  shareout_but(idx,"newuser %s\n",par);
  noshare=1; nsplit(etc,par); nsplit(etc2,par); nsplit(etc3,par);
  adduser(etc,etc2,etc3,atoi(par)); noshare=0;
  log(LOG_CMDS,"%s: newuser %s",dcc[idx].nick,etc);
}

void tand_killuser(idx,z,par)
int idx,z; char *par;
{
  CHKSHARE;
  shareout_but(idx,"killuser %s\n",par);
  log(LOG_CMDS,"%s: killuser %s",dcc[idx].nick,par);
  noshare=1; deluser(par); noshare=0;
}

void tand_pls_host(idx,z,par)
int idx,z; char *par;
{
  char hand[40];
  CHKSHARE;
  shareout_but(idx,"+host %s\n",par);
  noshare=1; nsplit(hand,par); addhost_by_handle(hand,par); noshare=0;
  log(LOG_CMDS,"%s: +host %s %s",dcc[idx].nick,hand,par);
}

void tand_mns_host(idx,z,par)
int idx,z; char *par;
{
  char hand[40];
  CHKSHARE;
  shareout_but(idx,"-host %s\n",par);
  noshare=1; nsplit(hand,par); delhost_by_handle(hand,par); noshare=0;
  log(LOG_CMDS,"%s: -host %s %s",dcc[idx].nick,hand,par);
}

void tand_chemail(idx,z,par)
int idx,z; char *par;
{
  char hand[40];
  CHKSHARE;
  shareout_but(idx,"chemail %s\n",par);
  noshare=1; nsplit(hand,par); set_handle_email(hand,par); noshare=0;
  log(LOG_CMDS,"%s: change email %s",dcc[idx].nick,hand);
}

void tand_chdccdir(idx,z,par)
int idx,z; char *par;
{
  char hand[40];
  CHKSHARE;
  shareout_but(idx,"chdccdir %s\n",par);
  noshare=1; nsplit(hand,par); set_handle_dccdir(hand,par); noshare=0;
}

void tand_chcomment(idx,z,par)
int idx,z; char *par;
{
  char hand[40];
  CHKSHARE;
  shareout_but(idx,"chcomment %s\n",par);
  noshare=1; nsplit(hand,par); set_handle_comment(hand,par); noshare=0;
  log(LOG_CMDS,"%s: change comment %s",dcc[idx].nick,hand);
}

void tand_chinfo(idx,z,par)
int idx,z; char *par;
{
  char hand[40];
  CHKSHARE;
  shareout_but(idx,"chinfo %s\n",par);
  noshare=1; nsplit(hand,par); set_handle_info(hand,par); noshare=0;
  log(LOG_CMDS,"%s: change info %s",dcc[idx].nick,hand);
}

void tand_pls_attr(idx,z,par)
int idx,z; char *par;
{
  char hand[40],attr[40];
  CHKSHARE;
  shareout_but(idx,"+attr %s\n",par);
  noshare=1; nsplit(hand,par); chg_attr_by_handle(hand,'+',atoi(par));
  strcpy(attr,attrname(atoi(par)));
  log(LOG_CMDS,"%s: +%s %s",dcc[idx].nick,attr,hand);
  noshare=0;
}

void tand_mns_attr(idx,z,par)
int idx,z; char *par;
{
  char hand[40],attr[40];
  CHKSHARE;
  shareout_but(idx,"-attr %s\n",par);
  noshare=1; nsplit(hand,par); chg_attr_by_handle(hand,'-',atoi(par));
  strcpy(attr,attrname(atoi(par)));
  log(LOG_CMDS,"%s: -%s %s",dcc[idx].nick,attr,hand);
  noshare=0;
}

void tand_mns_ban(idx,z,par)
int idx,z; char *par;
{
  CHKSHARE;
  shareout_but(idx,"-ban %s\n",par);
  log(LOG_CMDS,"%s: cancel ban %s",dcc[idx].nick,par);
  noshare=1; unprog_ban(par); noshare=0;
}

void tand_mns_ignore(idx,z,par)
int idx,z; char *par;
{
  CHKSHARE;
  shareout_but(idx,"-ignore %s\n",par);
  log(LOG_CMDS,"%s: cancel ignore %s",dcc[idx].nick,par);
  noshare=1; unprog_ignore(par); noshare=0;
}

void tand_pls_ban(idx,z,par)
int idx,z; char *par;
{
  time_t now=time(NULL); char ban[121],tm[40];
  CHKSHARE;
  shareout_but(idx,"+ban %s\n",par);
  noshare=1; nsplit(ban,par); nsplit(tm,par);
  log(LOG_CMDS,"%s: ban %s (%s)",dcc[idx].nick,ban,par);
  if (atol(tm)==0) prog_ban(ban,0,par);
  else prog_ban(ban,now-atol(tm),par);
  noshare=0;
}

void tand_pls_ignore(idx,z,par)
int idx,z; char *par;
{
  time_t now=time(NULL); char ign[121];
  CHKSHARE;
  shareout_but(idx,"+ignore %s\n",par);
  noshare=1; nsplit(ign,par);
  log(LOG_CMDS,"%s: ignore %s",dcc[idx].nick,ign);
  prog_ignore(ign,now-atoi(par)); noshare=0;
}

void tand_nlinked(idx,z,par)
int idx,z; char *par;
{
  char newbot[20],next[20];
  nsplit(newbot,par); nsplit(next,par);
  if ((in_chain(newbot)) || (strcasecmp(newbot,origbotname)==0)) {
    /* loop! */
    log(LOG_MISC,"Detected loop: disconnecting %s (mutual: %s)",dcc[idx].nick,
	newbot);
    chatout("*** Loop (%s): disconnected %s\n",newbot,dcc[idx].nick);
    tandout_but(idx,"chat %s Loop (%s): disconnected %s\n",origbotname,newbot,
		dcc[idx].nick);
    tandout_but(idx,"unlinked %s\n",dcc[idx].nick);
    cancel_user_xfer(idx);
    rembot(dcc[idx].nick,dcc[idx].nick); unvia(idx,dcc[idx].nick);
    tprintf(z,"bye\n");
    close(dcc[idx].sock); shutdown(dcc[idx].sock,2); lostdcc(idx);
  }
  else {
    addbot(newbot,dcc[idx].nick);
    got_tand_next(newbot,next);
    tandout_but(idx,"nlinked %s %s %s\n",newbot,next,par);
    if (match_attr_handle(dcc[idx].nick,BOT_LEAF)) {
      tprintf(dcc[idx].sock,"reject %s %s\n",origbotname,newbot);
      log(LOG_MISC,"Rejecting connection to %s by leaf %s",newbot,
	  dcc[idx].nick);
    }
    if (match_attr_handle(newbot,BOT_REJECT)) {
      tprintf(dcc[idx].sock,"reject %s %s\n",origbotname,newbot);
      log(LOG_MISC,"Rejecting bot %s from %s",newbot,dcc[idx].nick);
    }
  }
}

void tand_linked(idx,z,par)
int idx,z; char *par;
{
  if ((in_chain(par)) || (strcasecmp(par,origbotname)==0)) {
    /* loop! */
    log(LOG_MISC,"Detected loop: disconnecting %s (mutual: %s)",dcc[idx].nick,
	par);
    chatout("*** Loop (%s): disconnected %s\n",par,dcc[idx].nick);
    tandout_but(idx,"chat %s Loop (%s): disconnected %s\n",origbotname,
		par,dcc[idx].nick);
    tandout_but(idx,"unlinked %s\n",dcc[idx].nick);
    cancel_user_xfer(idx);
    rembot(dcc[idx].nick,dcc[idx].nick); unvia(idx,dcc[idx].nick);
    close(dcc[idx].sock); shutdown(dcc[idx].sock,2); lostdcc(idx);
  }
  else {
    addbot(par,dcc[idx].nick);
    tandout_but(idx,"linked %s\n",par);
    tprintf(dcc[idx].sock,"trace *@%s %s %s\n",origbotname,par,origbotname);
    if (match_attr_handle(dcc[idx].nick,BOT_LEAF)) {
      tprintf(dcc[idx].sock,"reject %s %s\n",origbotname,par);
      log(LOG_MISC,"Rejecting connection to %s by leaf %s",par,dcc[idx].nick);
    }
  }
}

void tand_unlinked(idx,z,par)
int idx,z; char *par;
{
  rembot(par,dcc[idx].nick);
  unvia(idx,par);
  tandout_but(idx,"unlinked %s\n",par);
}

void tand_trace(idx,z,par)
int idx,z; char *par;
{
  char from[40],dest[40]; int i;
  /* trace <from@bot> <dest> <chain:chain..> */
  nsplit(from,par); nsplit(dest,par);
  if (strcasecmp(dest,origbotname)==0) {
    tprintf(z,"traced %s %s:%s\n",from,par,origbotname);
  }
  else {
    i=nextbot(dest); if (i<0)
      tandout_but(idx,"trace %s %s %s:%s\n",from,dest,par,origbotname);
    else tprintf(dcc[i].sock,"trace %s %s %s:%s\n",from,dest,par,origbotname);
  }
}

void tand_traced(idx,z,par)
int idx,z; char *par;
{
  char *p,tobot[40],to[40],bot[40]; int i;
  /* traced <to@bot> <chain:chain..> */
  nsplit(tobot,par); splitc(to,tobot,'@');
  if (strcasecmp(tobot,origbotname)==0) {
    for (i=0; i<dcc_total; i++)
      if ((dcc[i].type==DCC_CHAT) && (strcasecmp(dcc[i].nick,to)==0))
	tprintf(dcc[i].sock,"Trace result -> %s\n",par);
    if (strcmp(to,"*")==0) {
      p=strrchr(par,':'); if (p!=NULL) {
	*p=0; strcpy(bot,p+1); p=strchr(par,':'); while (p!=NULL) {
	  p++; strcpy(par,p); p=strchr(par,':');
	}
	got_tand_next(bot,par);
      }
    }
  }
  else {
    i=nextbot(tobot); if (i<0)
      tandout_but(idx,"traced %s@%s %s\n",to,tobot,par);
    else tprintf(dcc[i].sock,"traced %s@%s %s\n",to,tobot,par);
  }
}

void tand_reject(idx,z,par)
int idx,z; char *par;
{
  char from[40]; int i;
  /* reject <from> <bot> */
  nsplit(from,par); if (strchr(par,'@')==NULL) {
    i=nextbot(par); if (i<0) {
      tprintf(z,"priv %s %s You can't reject that cos it doesn't exist.\n",
	      origbotname,from);
    }
    else if (strcasecmp(dcc[i].nick,par)==0) {
      /* i'm the connection to the rejected bot */
      log(LOG_MISC,"%s rejected %s",from,dcc[i].nick);
      tprintf(dcc[i].sock,"bye\n");
      tandout_but(i,"unlinked %s\n",dcc[i].nick);
      tandout_but(i,"chat %s Disconnected %s (rejected by %s)\n",origbotname,
		  dcc[i].nick,from);
      chatout("*** Disconnected %s (rejected by %s)\n",dcc[i].nick,from);
      cancel_user_xfer(i);
      rembot(dcc[i].nick,dcc[i].nick); unvia(i,dcc[i].nick);
      close(dcc[i].sock); shutdown(dcc[i].sock,2); lostdcc(i);
    }
    else {
      if (i<0) tandout_but(idx,"reject %s %s\n",from,par);
      else tprintf(dcc[i].sock,"reject %s %s\n",from,par);
    }
  }
  else {    /* rejecting user */
    tprintf(z,"priv %s %s Remote boots are not allowed.\n",origbotname,from);
  }
}

void tand_thisbot(idx,z,par)
int idx,z; char *par;
{
  if (strcasecmp(par,dcc[idx].nick)!=0) {
    log(LOG_MISC,"Wrong bot: wanted %s, got %s",dcc[idx].nick,par);
    tprintf(z,"bye\n");
    tandout_but(idx,"unlinked %s\n",dcc[idx].nick);
    tandout_but(idx,"chat %s Disconnected %s (imposter)\n",origbotname,
		dcc[idx].nick);
    chatout("*** Disconnected %s (imposter)\n",dcc[idx].nick);
    rembot(dcc[idx].nick,dcc[idx].nick); unvia(idx,dcc[idx].nick);
    close(dcc[idx].sock); shutdown(dcc[idx].sock,2); lostdcc(idx); return;
  }
  if (match_attr_handle(par,BOT_REJECT)) {
    log(LOG_MISC,"Rejecting bot %s",par);
    tprintf(z,"bye\n");
    tandout_but(idx,"unlinked %s\n",dcc[idx].nick);
    tandout_but(idx,"chat %s Disconnected %s (rejected)\n",origbotname,
		dcc[idx].nick);
    chatout("*** Disconnected %s (rejected)\n",dcc[idx].nick);
    rembot(dcc[idx].nick,dcc[idx].nick); unvia(idx,dcc[idx].nick);
    close(dcc[idx].sock); shutdown(dcc[idx].sock,2); lostdcc(idx); return;
  }
  /* set capitalization the way they want it */
  noshare=1; change_handle(dcc[idx].nick,par); noshare=0;
  strcpy(dcc[idx].nick,par);
}

void tand_handshake(idx,z,par)
int idx,z; char *par;
{
  change_pass_by_handle(dcc[idx].nick,par);
  if ((dcc[idx].u.tand->status&STAT_SHARE) &&
      (!(dcc[idx].u.tand->status&STAT_GETTING)) &&
      (!(dcc[idx].u.tand->status&STAT_SENDING))) {
    noshare=1; change_pass_by_handle(origbotname,par); noshare=0;
  }
}

void tand_trying(idx,z,par)
int idx,z; char *par;
{
  tandout_but(idx,"trying %s\n",par);
  /* currently ignore */
}

void tand_end_trying(idx,z,par)
int idx,z; char *par;
{
  tandout_but(idx,"*trying %s\n",par);
  /* currently ignore */
}

/* make sure another bot's password = this bot's password */
void tand_sharing(idx,z,par)
int idx,z; char *par;
{
  char s[40];
/* ignore a 'sharing' command now (9g) */
/*  get_pass_by_handle(dcc[idx].nick,s);
  if (!(pass_match_by_handle(s,par)))
    change_pass_by_handle(par,s); */
}

/* possibly op another bot */
void tand_op(idx,z,par)
int idx,z; char *par;
{
  if (match_attr_handle(dcc[idx].nick,USER_OP))
    add_mode('+','o',par);
}
